import vulnerability
import data


def get_new_key(d):
    if not d.keys():
        return 0
    return max(d.keys()) + 1


def get_feature(asset, ID = None):
    result = asset.__geo_interface__
    if ID is not None:
        result.update({'id':ID})
    return result
    
      


    shapely.geometry.mapping(asset.geometry)


def get_collection(assets):
    features = []
    for ID, a in assets.items():
        features.append(get_feature(a, ID))

    return {"type": "FeatureCollection","features": features}
    


#possible approach: make sure this is still usable without a web interface
#ie - do conversions elsewhere

class Project(object):

    def __init__(self):

        self.vulnerabilities = {}
        self.datasources = {}
        self.assets = {}
        self.results = {}

        self.dtypes = {}

    def newCalculation(self, did, vid):
        self.results[(did, vid)] = None

    def updateCalculation(self, old_did, old_vid, new_did, new_vid):
        if (old_did, old_vid) == (new_did, new_vid):
            return
        if (old_did, old_vid) in self.results:
            self.results.pop((old_did, old_vid))
        self.results[(new_did, new_vid)] = None


    def getResults(self, did, vid):
        if self.results[(did, vid)] is None:
            print "calculating new results"
            self.results[(did, vid)] = self._calculate_results(did, vid)
        else:
            print "using old results"
        return self.results[(did, vid)]


    def getAssets(self, did):
        if self.assets.get(did) is None:
            self.assets[did] = self._get_assets(did)
        return self.assets[did]


    def _discard(self, changed_did = None, changed_vid = None):
        if changed_vid is not None:
            for vid, did in self.results:
                if vid == changed_vid:
                    self.results[(vid, did)] = None

        if changed_did is not None:
            for vid, did in self.results:
                if did == changed_did:
                    self.results[(vid, did)] = None
                    self.assets[did] = None

    def _get_assets(self, did):
        d = self.datasources[did]
        crs, assets = d.parse()
        return crs, assets
        


    def _calculate_results(self, did, vid):
        '''run through all assets, build dict of results'''
        crs, assets = self.getAssets(did)
        v = self.vulnerabilities[vid]


        scores = {}

        for ID, a in assets.items():
            scores[ID] = v.getVulnerability(a)

        return scores



    def tabulate(self, dsid, sort_vid = None, vids = None):

        if vids is None:
            vids = filter(lambda x: x[0] == dsid, self.results.keys())
            vids = zip(*vids)[1]


        #two layer key: asset_id, then vulnerability id
        scores = {}

        for i in vids:
            sco = self.getResults(dsid, i)

            
            for asset_id, s in sco.items():

                scores.setdefault(asset_id, {})
                scores[asset_id][i] = s

        scores = scores.items()

        if sort_vid is not None:
            scores.sort(key = lambda x: x[1][sort_vid].overall, reverse = True)
        else:
            scores.sort(key = lambda x: x[0])

        return scores, vids
            
            



    def newVulnerability(self):
        v = vulnerability.Vulnerability(
            '',
            dict.fromkeys(vulnerability.Vulnerability.components, 1.0),
            dict([(x, []) for x in vulnerability.Vulnerability.components]))

        ID = get_new_key(self.vulnerabilities)

        self.vulnerabilities[ID] = v

        return ID

    def updateVulnerability(self, vid, name, weights):
        #here we just update attributes
        #this circumvents the initialization validation
        #maybe move validation outside of initialization?

        change = False

        v = self.vulnerabilities[vid]
        v.name = name

        if weights != v.weights:
            change = True

        v.weights = weights

        if change:
            self._discard(changed_vid = vid)

            


    def deleteVulnerability(self, vid):
        pass

    def newIndicator(self, vid, component):

        #an issue to take care of for indicators is data types
        #not an issure for name, weight, field, x, y or mapping values
        #but is an issue for missing and mapping keys

        #missing is for indicators
        #mapping is for normalizer

        #maybe indicators can have a data type assigned to them in the project
        #that data type is used to interpret missing values and mapping keys
        #as they are read from the http request

        vid = int(vid)
        

        weight = 1.0

        indicator = vulnerability.Field('','')

        normalizer = vulnerability.PiecewiseLinearInterpolation(
            x = [1,2,3,4])

        v = self.vulnerabilities[vid]

        v.indicators[component].append((weight, indicator, normalizer))

        index = len(v.indicators[component]) - 1

        self.dtypes[(vid, component, index)] = 'int'

        return index

    @staticmethod
    def _get_normalizer_type(normalizer):
        if normalizer.__class__.__name__ == 'DiscreteMapping':
            return 'discrete'
        else:
            return 'interpolate'

    

    def updateIndicator(self, vid, component, index,
                        name, weight, field, missing, type_, dtype,
                        x = None, y = None, mapping = None):


        change = False

        v = self.vulnerabilities[vid]

        old_weight, old_indicator, old_normalizer = v.indicators[component][index]

        old_dtype = self.dtypes[(vid, component, index)]
        
        if abs(weight - old_weight) > 1e-9:
            change = True

        if old_dtype != dtype:
            change = True

        if old_indicator.field != field:
            change = True

        if old_indicator.missing != missing:
            change = True

        old_indicator.field = field
        old_indicator.missing = missing
        old_indicator.name = name





        old_type = self._get_normalizer_type(old_normalizer)


        if type_ != old_type:
            change = True
            if type_ == 'discrete':
                new_normalizer = vulnerability.DiscreteMapping(mapping)

            else:
                new_normalizer = vulnerability.PiecewiseLinearInterpolation(x,y)

            result = (weight, old_indicator, new_normalizer)

        else:
            if type_ == 'discrete':
                if old_normalizer.mapping != mapping:
                    change = True
                old_normalizer.mapping = mapping
            else:
                
                
                if old_normalizer.x != sorted(x):
                    change = True
                if old_normalizer.y != sorted(y):
                    change = True
                old_normalizer._set_xy(x,y)

            result = (weight, old_indicator, old_normalizer)

        v.indicators[component][index] = result

        self.dtypes[(vid, component, index)] = dtype

        if change:
            self._discard(changed_vid = vid)

                    

    def deleteIndicator(self, component, index):
        pass


    def newDataSource(self):
        d = data.ShapeFile('', '', '', '')
        ID = get_new_key(self.datasources)

        self.datasources[ID] = d

        return ID

    def updateDataSource(self, ID, name, path, id_field, name_field):
        change = False

        #data sources need a way to communicate changes on disk
        d = self.datasources[ID]

        if d.path != path:
            change = True

        if d.id_field != id_field:
            change = True

        new_datasource = data.ShapeFile(name, path, id_field, name_field)

        self.datasources[ID] = new_datasource

        if change:
            self._discard(changed_did = ID)
        
        

    def deleteDataSource(self):
        pass
    

        

        

        
 

    


    


        
            



    



        



  
